from typing import Generic, Any, TypeVar, Optional, Dict, Union

from unipipeline.message.uni_message import UniMessage
from unipipeline.worker.uni_worker_consumer_manager import UniWorkerConsumerManager
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage

TInputMessage = TypeVar('TInputMessage', bound=UniMessage)
TOutputMessage = TypeVar('TOutputMessage', bound=Optional[UniMessage])


class UniWorker(Generic[TInputMessage, TOutputMessage]):
    def __init__(self, manager: UniWorkerConsumerManager) -> None:
        self._uni_manager = manager

    @property
    def manager(self) -> UniWorkerConsumerManager:
        return self._uni_manager

    def handle_message(self, msg: UniWorkerConsumerMessage[TInputMessage]) -> Optional[Union[TOutputMessage, Dict[str, Any]]]:
        raise NotImplementedError(f'method handle_message not implemented for {type(self).__name__}')



from datetime import datetime

from example.messages.ender_after_input_message import EnderAfterInputMessage
from example.messages.input_message import InputMessage
from example.messages.some_external_message import SomeExternalMessage
from example.workers.ender_after_input_worker import EnderAfterInputWorker
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage


class InputWorker(UniWorker[InputMessage, None]):
    def handle_message(self, msg: UniWorkerConsumerMessage[InputMessage]) -> None:
        # answ = '!!!'

        answ = self.manager.get_answer_from(EnderAfterInputWorker, EnderAfterInputMessage(
            value=f'from input_worker {datetime.now()}'
        ))

        self.manager.send_to('some_external_worker', SomeExternalMessage(
            value=f'answ: {answ} ==> from input_worker {datetime.now()}'
        ))



from typing import Callable

from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager


class UniKafkaBrokerMessageManager(UniBrokerMessageManager):
    def __init__(self, commit: Callable[[], None]) -> None:
        self._commit = commit
        self._acknowledged = False

    def reject(self) -> None:
        pass

    def ack(self) -> None:
        if self._acknowledged:
            return
        self._acknowledged = True
        self._commit()



class UniBrokerMessageManager:
    def reject(self) -> None:
        raise NotImplementedError(f'method reject must be specified for class "{type(self).__name__}"')

    def ack(self) -> None:
        raise NotImplementedError(f'method acknowledge must be specified for class "{type(self).__name__}"')



from time import sleep, time
from typing import Dict, TypeVar, Any, Set, Union, Optional, Type, List, NamedTuple, Callable
from uuid import uuid4

from unipipeline.errors.uni_payload_error import UniPayloadSerializationError
from unipipeline.errors.uni_work_flow_error import UniWorkFlowError
from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.brokers.uni_broker import UniBroker, UniBrokerConsumer
from unipipeline.config.uni_config import UniConfig
from unipipeline.modules.uni_cron_job import UniCronJob
from unipipeline.utils.uni_echo import UniEcho
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniMessageMetaErrTopic, UniAnswerParams
from unipipeline.utils.uni_util import UniUtil
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer import UniWorkerConsumer
from unipipeline.definitions.uni_worker_definition import UniWorkerDefinition
from unipipeline.utils.sig import soft_interruption

TWorker = TypeVar('TWorker', bound=UniWorker[Any, Any])


class UniBrokerInitRecipe(NamedTuple):
    topics: Set[str]
    answer_topics: Set[str]


class UniMediator:
    def __init__(self, util: UniUtil, echo: UniEcho, config: UniConfig) -> None:
        self._config = config
        self._echo = echo
        self._util = util

        self._worker_definition_by_type: Dict[Any, UniWorkerDefinition] = dict()
        self._worker_instance_indexes: Dict[str, UniWorkerConsumer[Any, Any]] = dict()
        self._broker_instance_indexes: Dict[str, UniBroker[Any]] = dict()
        self._worker_init_list: Set[str] = set()
        self._worker_initialized_list: Set[str] = set()
        self._waiting_init_list: Set[str] = set()
        self._waiting_initialized_list: Set[str] = set()

        self._consumers_list: Set[str] = set()
        self._brokers_with_topics_to_init: Dict[str, UniBrokerInitRecipe] = dict()
        self._brokers_with_topics_initialized: Dict[str, UniBrokerInitRecipe] = dict()

        self._message_types: Dict[str, Type[UniMessage]] = dict()

        self._connected_brokers: List[UniBroker[Any]] = list()
        self._brokers_with_active_consumption: List[UniBroker[Any]] = list()

        self._decompression_modules: Dict[str, Callable[[bytes], bytes]] = dict()
        self._compression_modules: Dict[str, Callable[[bytes], bytes]] = dict()
        self._content_type_serializers: Dict[str, Callable[[Dict[str, Any]], Union[str, bytes]]] = dict()
        self._content_type_parsers: Dict[str, Callable[[Union[str, bytes]], Dict[str, Any]]] = dict()

    @property
    def echo(self) -> UniEcho:
        return self._echo

    def set_echo_level(self, level: int) -> None:
        self.echo.level = level

    def get_broker(self, name: str, singleton: bool = True) -> UniBroker[Any]:
        if not singleton:
            broker_def = self.config.brokers[name]
            broker_type = broker_def.type.import_class(UniBroker, self.echo, util=self._util)
            br = broker_type(mediator=self, definition=broker_def)
            return br
        if name not in self._broker_instance_indexes:
            self._broker_instance_indexes[name] = self.get_broker(name, singleton=False)
        return self._broker_instance_indexes[name]

    def move_to_error_topic(self, wd: UniWorkerDefinition, meta: UniMessageMeta, err_topic: UniMessageMetaErrTopic, err: Exception) -> None:
        self._echo.log_error(str(err))
        meta = meta.create_error_child(err_topic, err)
        br = self.get_broker(wd.broker.name)
        error_topic = wd.error_topic
        if error_topic == UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR.value:
            error_topic = wd.error_payload_topic
        br.publish(error_topic, [meta])

    def add_worker_to_consume_list(self, name: str) -> None:
        wd = self._config.workers[name]
        if wd.marked_as_external:
            raise OverflowError(f'your could not use worker "{name}" as consumer. it marked as external "{wd.external}"')
        self._consumers_list.add(name)
        self.echo.log_info(f'added consumer {name}')

    def get_message_type(self, name: str) -> Type[UniMessage]:
        if name in self._message_types:
            return self._message_types[name]

        self._message_types[name] = self.config.messages[name].type.import_class(UniMessage, self.echo, util=self._util)

        return self._message_types[name]

    def answer_to(self, worker_name: str, req_meta: UniMessageMeta, payload: Optional[Union[Dict[str, Any], UniMessageMeta, UniMessage]], *, unwrapped: bool) -> None:
        wd = self._config.workers[worker_name]
        if not wd.need_answer:
            if payload is not None:
                raise UniWorkFlowError(f'output message must be None because worker {wd.name} has no possibility to send output messages')
            return

        if payload is None:
            raise UniPayloadSerializationError('output message must be not empty')

        answ_meta: UniMessageMeta
        if isinstance(payload, UniMessageMeta):
            answ_meta = payload
        else:
            assert wd.answer_message is not None
            answ_message_type = self.get_message_type(wd.answer_message.name)
            payload_msg: UniMessage
            if isinstance(payload, answ_message_type):
                payload_msg = payload
            elif isinstance(payload, dict):
                try:
                    payload_msg = answ_message_type(**payload)  # type: ignore
                except Exception as e:  # noqa
                    raise UniPayloadSerializationError(str(e))
            else:
                raise UniPayloadSerializationError(f'output message has invalid type. {type(payload).__name__} was given')

            answ_meta = req_meta.create_child(payload_msg.dict(), unwrapped=unwrapped)

        b = self.get_broker(wd.broker.name)

        assert req_meta.answer_params is not None
        b.publish_answer(req_meta.answer_params, answ_meta)
        self.echo.log_info(f'worker {worker_name} answers to {req_meta.answer_params.topic}->{req_meta.answer_params.id} :: {answ_meta}')

    def send_to(
        self,
        worker_name: str,
        payload: Union[Dict[str, Any], UniMessage],
        *,
        parent_meta: Optional[UniMessageMeta] = None,
        answer_params: Optional[UniAnswerParams] = None,
        alone: bool = False,
    ) -> Optional[UniAnswerMessage[UniMessage]]:
        if worker_name not in self._worker_initialized_list:
            raise OverflowError(f'worker {worker_name} was not initialized')

        wd = self._config.workers[worker_name]

        message_type = self.get_message_type(wd.input_message.name)
        try:
            if isinstance(payload, message_type):
                payload_data = payload.dict()
            elif isinstance(payload, dict):
                payload_data = message_type(**payload).dict()  # type: ignore
            else:
                raise TypeError(f'data has invalid type.{type(payload).__name__} was given')
        except Exception as e:  # noqa
            raise UniPayloadSerializationError(str(e))

        br = self.get_broker(wd.broker.name)

        if parent_meta is not None:
            meta = parent_meta.create_child(payload_data, unwrapped=wd.input_unwrapped, answer_params=answer_params)
        else:
            meta = UniMessageMeta.create_new(payload_data, unwrapped=wd.input_unwrapped, answer_params=answer_params)

        if meta.need_answer and wd.need_answer:
            assert wd.answer_message is not None
            answ_meta = br.rpc_call(wd.topic, meta, alone=alone, max_delay_s=wd.answer_avg_delay_s * 3, unwrapped=wd.answer_unwrapped)
            if answ_meta is None:
                if alone:
                    return None
                raise OverflowError('system error. answer object must be specified')
            return UniAnswerMessage(answ_meta, self.get_message_type(wd.answer_message.name))

        meta_list = [meta]
        br.publish(wd.topic, meta_list, alone=alone)  # TODO: make it list by default
        self.echo.log_info(f"worker {wd.name} sent message to topic '{wd.topic}':: {meta_list}")
        return None

    def start_cron(self) -> None:
        cron_jobs = UniCronJob.mk_jobs_list(self.config.cron_tasks.values(), self)
        self.echo.log_debug(f'cron jobs defined: {", ".join(cj.task.name for cj in cron_jobs)}')

        prev_time = time()
        while True:
            now = time()
            if (now - prev_time) < 0.8:
                sleep(.1)  # delay for correct next iteration
                continue
            delay, jobs = UniCronJob.search_next_tasks(cron_jobs)
            if delay is None:
                self.echo.log_warning("no active cron tasks found")
                return
            self.echo.log_debug(f"sleep {delay} seconds before running the tasks: {[cj.task.name for cj in jobs]}")
            if delay > 0:
                sleep(delay)
            self.echo.log_info(f"run the tasks: {[cj.task.name for cj in jobs]}")
            prev_time = time()
            for cj in jobs:
                cj.send()

    def exit(self) -> None:
        for b in self._connected_brokers:
            b.close()
        self._connected_brokers = list()

    def start_consuming(self) -> None:
        brokers = set()
        for wn in self._consumers_list:
            wd = self._config.workers[wn]
            wc = self.get_worker_consumer(wn)

            br = self.get_broker(wd.broker.name)

            self.echo.log_info(f"worker {wn} start consuming")
            br.add_consumer(UniBrokerConsumer(
                topic=wd.topic,
                id=f'{wn}__{uuid4()}',
                group_id=wn,
                unwrapped=wd.input_unwrapped,
                message_handler=wc.process_message,
                prefetch_count=wd.prefetch_count,
            ))

            self.echo.log_info(f'consumer {wn} initialized')
            brokers.add(wd.broker.name)

        with soft_interruption(self._handle_interruption, self._handle_force_interruption, self._interruption_err):
            for bn in brokers:
                b = self.get_broker(bn)
                self._brokers_with_active_consumption.append(b)
                self.echo.log_info(f'broker {bn} consuming start')
                b.start_consuming()

    def _interruption_err(self, err: Exception) -> None:
        self.echo.log_error(str(err))
        raise err

    def _handle_force_interruption(self) -> None:
        self.echo.log_warning('force interruption detected')

    def _handle_interruption(self) -> None:
        self.echo.log_warning('interruption detected')
        for b in self._brokers_with_active_consumption:
            self.echo.log_debug(f'broker "{b.definition.name}" was notified about interruption')
            b.stop_consuming()
        self._brokers_with_active_consumption = list()
        self.echo.log_info('all brokers was notified about interruption')

    def get_compressor(self, name: str) -> Callable[[bytes], bytes]:
        if name in self._compression_modules:
            return self._compression_modules[name]
        if name in self._config.compression:
            c = self._config.compression[name]
            self._compression_modules[name] = c.encoder_type.import_function()
            return self._compression_modules[name]
        raise ValueError(f'compression of "{name}" is not supported')

    def get_decompressor(self, name: str) -> Callable[[bytes], bytes]:
        if name in self._decompression_modules:
            return self._decompression_modules[name]
        if name in self._config.compression:
            c = self._config.compression[name]
            self._decompression_modules[name] = c.decoder_type.import_function()
            return self._decompression_modules[name]
        raise ValueError(f'decompression of "{name}" is not supported')

    def decompress_message_body(self, compression: Optional[str], data: Union[str, bytes]) -> bytes:
        data_bytes: bytes
        if isinstance(data, str):
            data_bytes = bytes(data, encoding='utf-8')
        elif isinstance(data, bytes):
            data_bytes = data
        else:
            raise TypeError('invalid type')
        if compression is not None:
            decompressor = self.get_decompressor(compression)
            return decompressor(data_bytes)
        return data_bytes

    def compress_message_body(self, compression: Optional[str], data: Union[str, bytes]) -> bytes:
        data_bytes: bytes
        if isinstance(data, str):
            data_bytes = bytes(data, encoding='utf-8')
        elif isinstance(data, bytes):
            data_bytes = data
        else:
            raise TypeError('invalid type')
        if compression is not None:
            compressor = self.get_compressor(compression)
            return compressor(data_bytes)
        return data_bytes

    def parse_content_type(self, content_type: str, data: Union[bytes, str]) -> Dict[str, Any]:
        data_str: str
        if isinstance(data, str):
            data_str = data
        elif isinstance(data, bytes):
            data_str = data.decode('utf-8')
        else:
            raise TypeError('invalid type')

        parser = self.get_content_type_parser(content_type)
        return parser(data_str)

    def serialize_content_type(self, content_type: str, data: Dict[str, Any]) -> bytes:
        if not isinstance(data, dict):
            raise TypeError(f'invalid type of payload. must be dict, {type(data).__name__} was given')

        serializer = self.get_content_type_serializer(content_type)
        res = serializer(data)
        if isinstance(res, str):
            return res.encode('utf-8')
        return res

    def get_content_type_serializer(self, name: str) -> Callable[[Dict[str, Any]], Union[bytes, str]]:
        if name in self._content_type_serializers:
            return self._content_type_serializers[name]
        if name in self._config.codecs:
            c = self._config.codecs[name]
            self._content_type_serializers[name] = c.encoder_type.import_function()
            return self._content_type_serializers[name]
        raise ValueError(f'content_type "{name}" is not supported')

    def get_content_type_parser(self, name: str) -> Callable[[Union[bytes, str]], Dict[str, Any]]:
        if name in self._content_type_parsers:
            return self._content_type_parsers[name]
        if name in self._config.codecs:
            c = self._config.codecs[name]
            self._content_type_parsers[name] = c.decoder_type.import_function()
            return self._content_type_parsers[name]
        raise ValueError(f'content_type "{name}" is not supported')

    def add_worker_to_init_list(self, name: str, no_related: bool) -> None:
        if name not in self._config.workers:
            self.echo.exit_with_error(f'worker "{name}" is not found in config "{self.config.file}"')
        wd = self._config.workers[name]
        self._worker_init_list.add(name)
        for waiting in wd.waitings:
            if waiting.name not in self._waiting_initialized_list:
                self._waiting_init_list.add(waiting.name)
        self.add_broker_topic_to_init(wd.broker.name, wd.topic, False)
        self.add_broker_topic_to_init(wd.broker.name, wd.error_topic, False)
        self.add_broker_topic_to_init(wd.broker.name, wd.error_payload_topic, False)
        if wd.need_answer:
            self.add_broker_topic_to_init(wd.broker.name, wd.answer_topic, True)
        if not no_related:
            for wn in wd.output_workers:
                self._worker_init_list.add(wn)
                owd = self._config.workers[wn]
                self.add_broker_topic_to_init(owd.broker.name, owd.topic, False)

    def add_broker_topic_to_init(self, name: str, topic: str, is_answer: bool) -> None:
        if name in self._brokers_with_topics_initialized:
            if is_answer:
                if topic in self._brokers_with_topics_initialized[name].answer_topics:
                    return
            else:
                if topic in self._brokers_with_topics_initialized[name].topics:
                    return

        if name not in self._brokers_with_topics_to_init:
            self._brokers_with_topics_to_init[name] = UniBrokerInitRecipe(set(), set())

        if is_answer:
            self._brokers_with_topics_to_init[name].answer_topics.add(topic)
        else:
            self._brokers_with_topics_to_init[name].topics.add(topic)

    def initialize(self, create: bool = True) -> None:
        echo = self.echo.mk_child('initialize')
        for wn in self._worker_init_list:
            echo.log_info(f'worker "{wn}"', )
            self._worker_initialized_list.add(wn)
        self._worker_init_list = set()

        for waiting_name in self._waiting_init_list:
            self._config.waitings[waiting_name].wait(echo)
            echo.log_info(f'waiting "{waiting_name}"')
            self._waiting_initialized_list.add(waiting_name)
        self._waiting_init_list = set()

        if create:
            for bn, collection in self._brokers_with_topics_to_init.items():
                bd = self._config.brokers[bn]

                if bd.marked_as_external:
                    echo.log_debug(f'broker "{bn}" skipped because it external')
                    continue

                b = self.wait_for_broker_connection(bn)

                b.initialize(collection.topics, collection.answer_topics)
                echo.log_info(f'broker "{b.definition.name}" topics :: {collection.topics}')
                if len(collection.answer_topics) > 0:
                    echo.log_info(f'broker "{b.definition.name}" answer topics :: {collection.answer_topics}')

                if bn not in self._brokers_with_topics_initialized:
                    self._brokers_with_topics_initialized[bn] = UniBrokerInitRecipe(set(), set())
                for topic in collection.topics:
                    self._brokers_with_topics_initialized[bn].topics.add(topic)
                for topic in collection.answer_topics:
                    self._brokers_with_topics_initialized[bn].answer_topics.add(topic)
            self._brokers_with_topics_to_init = dict()

    def get_worker_consumer(self, worker: Union[Type['UniWorker[Any, Any]'], str], *, singleton: bool = True) -> UniWorkerConsumer[Any, Any]:
        wd = self._config.get_worker_definition(worker)
        if wd.marked_as_external:
            raise OverflowError(f'worker "{worker}" is external. you could not get it')
        if not singleton or wd.name not in self._worker_instance_indexes:
            assert wd.type is not None
            worker_type = wd.type.import_class(UniWorker, self.echo, util=self._util)
            self.echo.log_info(f'get_worker :: initialized worker "{wd.name}"')
            wc = UniWorkerConsumer(wd, self, worker_type)
        else:
            return self._worker_instance_indexes[wd.name]
        self._worker_instance_indexes[wd.name] = wc
        return wc

    @property
    def config(self) -> UniConfig:
        return self._config

    def wait_for_broker_connection(self, name: str) -> UniBroker[Any]:
        br = self.get_broker(name)
        for try_count in range(br.definition.retry_max_count):
            try:
                br.connect()
                self.echo.log_info(f'wait_for_broker_connection :: broker {br.definition.name} connected')
                self._connected_brokers.append(br)
                return br
            except ConnectionError as e:
                self.echo.log_info(f'wait_for_broker_connection :: broker {br.definition.name} retry to connect [{try_count}/{br.definition.retry_max_count}] : {e}')
                sleep(br.definition.retry_delay_s)
                continue
        raise ConnectionError(f'unavailable connection to {br.definition.name}')



import logging
import sys
from typing import IO, Union, Optional

from unipipeline.utils.uni_util_color import UniUtilColor

SUPPORTED_LVL = {logging.INFO, logging.DEBUG, logging.WARNING, logging.ERROR}


def get_lvl(lvl: Union[str, int]) -> int:
    if isinstance(lvl, str):
        lvl = lvl.strip().lower()
        if lvl == 'info':
            return logging.INFO
        if lvl == 'debug':
            return logging.DEBUG
        if lvl == 'error':
            return logging.ERROR
        if lvl == 'warning':
            return logging.WARNING
        raise ValueError('invalid value of level')
    assert lvl in SUPPORTED_LVL
    return lvl


class UniEcho:
    def __init__(self, name: str, colors: UniUtilColor, prefix: str = '', level: Optional[Union[int, str]] = None) -> None:
        self._level_set = level is not None
        self._name = name
        self._level = get_lvl(level if level is not None else 'info')
        self._colors = colors

        prefix = f'{f"{prefix} | " if prefix else ""}{self._name}'
        self._debug_prefix = self._colors.color_it(self._colors.COLOR_GRAY, f'{prefix} | DEBUG   :: ')
        self._info_prefix = self._colors.color_it(self._colors.COLOR_CYAN, f'{prefix} | INFO    :: ')
        self._warn_prefix = self._colors.color_it(self._colors.COLOR_YELLOW, f'{prefix} | WARNING :: ')
        self._err_prefix = self._colors.color_it(self._colors.COLOR_RED, f'{prefix} | ERROR   :: ')
        self._success_prefix = self._colors.color_it(self._colors.COLOR_GREEN, f'{prefix} :: ')

        self._prefix = prefix

    def mk_child(self, name: str) -> 'UniEcho':
        e = UniEcho(name, colors=self._colors, prefix=self._prefix, level=self._level)
        return e

    @property
    def level(self) -> int:
        return self._level

    @level.setter
    def level(self, value: Union[int, str]) -> None:
        if self._level_set:
            return
        self._level = get_lvl(value)

    def echo(self, msg: str, stream: IO[str] = sys.stdout) -> None:
        stream.write(f'{msg}\n')

    def log_debug(self, msg: str) -> None:
        if logging.DEBUG >= self._level:
            self.echo(f'{self._debug_prefix}{msg}')

    def log_info(self, msg: str) -> None:
        if logging.INFO >= self._level:
            self.echo(f'{self._info_prefix}{msg}')

    def log_warning(self, msg: str) -> None:
        if logging.WARNING >= self._level:
            self.echo(f'{self._warn_prefix}{msg}', stream=sys.stderr)

    def log_error(self, msg: str) -> None:
        if logging.ERROR >= self._level:
            self.echo(f'{self._err_prefix}{msg}', stream=sys.stderr)

    def exit_with_error(self, msg: str) -> None:
        self.echo(f'{self._err_prefix}{msg}', stream=sys.stderr)
        exit(1)

    def success(self, msg: str) -> None:
        self.echo(f'{self._success_prefix}{self._colors.color_it(self._colors.COLOR_GREEN, msg)}')



from typing import TYPE_CHECKING

from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager

if TYPE_CHECKING:
    from unipipeline.brokers.uni_memory_broker_queue import UniMemoryBrokerQueue


class UniMemoryBrokerMessageManager(UniBrokerMessageManager):
    def __init__(self, ql: 'UniMemoryBrokerQueue', msg_id: int) -> None:
        self._msg_id = msg_id
        self._ql = ql

    def reject(self) -> None:
        self._ql.move_back_from_reserved(self._msg_id)

    def ack(self) -> None:
        self._ql.mark_as_processed(self._msg_id)



import functools
import threading
import time
import urllib.parse
from time import sleep
from typing import Optional, TypeVar, Set, List, NamedTuple, Callable, TYPE_CHECKING, Dict, Tuple, Any, cast, Type
from urllib.parse import urlparse

import amqp  # type: ignore
from amqp.exceptions import ConnectionError as AqmpConnectionError, RecoverableChannelError, AMQPError, ConnectionForced  # type: ignore

from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.brokers.uni_broker_message_manager import UniBrokerMessageManager
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_definition import UniDynamicDefinition
from unipipeline.errors.uni_answer_delay_error import UniAnswerDelayError
from unipipeline.message.uni_message import UniMessage
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniAnswerParams
from unipipeline.utils.uni_echo import UniEcho

import logging
logging.getLogger('amqp').setLevel(logging.DEBUG)
logging.getLogger('amqp.connection.Connection.heartbeat_tick').setLevel(logging.DEBUG)
logging.basicConfig(level=logging.DEBUG)

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator

BASIC_PROPERTIES__HEADER__COMPRESSION_KEY = 'compression'

RECOVERABLE_ERRORS = tuple({AqmpConnectionError, RecoverableChannelError, *amqp.Connection.recoverable_connection_errors})


TMessage = TypeVar('TMessage', bound=UniMessage)

# logging.getLogger('amqp').setLevel(logging.DEBUG)

T = TypeVar('T')


class UniPyPikaBrokerMessageManager(UniBrokerMessageManager):

    def __init__(self, channel: amqp.Channel, delivery_tag: str, reject: Callable[[amqp.Channel, str], None], ack: Callable[[amqp.Channel, str], None]) -> None:
        self._channel = channel
        self._delivery_tag = delivery_tag
        self._acknowledged = False
        self._reject = reject
        self._ack = ack

    def reject(self) -> None:
        self._reject(self._channel, self._delivery_tag)

    def ack(self) -> None:
        if self._acknowledged:
            return
        self._acknowledged = True
        self._ack(self._channel, self._delivery_tag)


class UniAmqpPyBrokerConfig(UniDynamicDefinition):
    exchange_name: str = "communication"
    answer_exchange_name: str = "communication_answer"
    heartbeat: int = 60
    prefetch: int = 1
    retry_max_count: int = 100
    retry_delay_s: int = 1
    persistent_message: bool = True

    mandatory_publishing: bool = False


class UniAmqpPyBrokerMsgProps(NamedTuple):
    content_type: Optional[str] = None
    content_encoding: Optional[str] = None
    application_headers: Optional[Dict[str, str]] = None
    delivery_mode: Optional[int] = None
    priority: Optional[int] = None
    correlation_id: Optional[str] = None
    reply_to: Optional[str] = None
    expiration: Optional[str] = None
    message_id: Optional[str] = None
    timestamp: Optional[int] = None
    type: Optional[str] = None
    user_id: Optional[str] = None
    app_id: Optional[str] = None
    cluster_id: Optional[str] = None


class UniAmqpPyBrokerConsumer(NamedTuple):
    queue: str
    on_message_callback: Callable[[amqp.Channel, 'amqp.Message'], None]
    consumer_tag: str
    prefetch_count: int


TFn = TypeVar('TFn', bound=Callable[..., Any])


def retryable(
    fn: TFn,
    echo: UniEcho,
    *,
    retryable_errors: Tuple[Type[Exception], ...],
    on_retries_ends: Optional[Callable[[], None]] = None,
    retry_max_count: int = 3,
    retry_delay_s: int = 1,
) -> TFn:
    echo = echo.mk_child(fn.__name__)
    max_retries = max(retry_max_count, 1)
    retry_threshold_s = retry_delay_s * max_retries
    assert retry_delay_s >= 0

    @functools.wraps(fn)
    def fn_wrapper(*args: Any, **kwargs: Any) -> Any:
        retry_counter = 0
        while True:
            starts_at = time.time()
            try:
                res = fn(*args, **kwargs)
                retry_counter = 0
                return res
            except retryable_errors as e:
                echo.log_warning(f'found error :: {e}')
                if int(time.time() - starts_at) >= retry_threshold_s:
                    retry_counter = 0
                if retry_counter >= max_retries:
                    echo.log_error(f'max retries {max_retries} was reached :: {e}')
                    if on_retries_ends is not None:
                        return on_retries_ends()
                    raise
                retry_counter += 1
                if retry_delay_s > 0:
                    sleep(retry_delay_s)
                echo.log_warning(f'retry {retry_counter}/{max_retries} :: {e}')
    return cast(TFn, fn_wrapper)


class AmqpPyChannelObj:
    def __init__(self, name: str, ttl: int, broker: 'UniAmqpPyBroker') -> None:
        self._name = name
        self._broker = broker
        self._ch: Optional[amqp.Channel] = None
        self._ttl = ttl
        self._ch_time: float = 0

    def close_channel(self) -> None:
        if self._ch is None or not self._ch.is_open:
            return
        self._ch.close()
        self._broker.echo.log_info(f'channel {self._name} closed')

    def get_channel(self, *, force_recreate: bool = False, force_current: bool = False) -> amqp.Channel:
        if force_recreate and not force_current:
            self.close_channel()
            return self.get_channel(force_recreate=False)

        if self._ch is None:
            self._ch_time = time.time()
            self._ch = self._broker.connected_connection.channel()
        elif not self._ch.is_open:
            self._ch.open()
            self._ch_time = time.time()
            self._broker.echo.log_info(f'channel {self._name} established')
        else:
            if not force_current and (time.time() - self._ch_time > self._ttl):
                return self.get_channel(force_recreate=True)
            self._broker.echo.log_debug(f'channel {self._name} reused')
        return self._ch


class UniAmqpPyBroker(UniBroker[UniAmqpPyBrokerConfig]):
    config_type = UniAmqpPyBrokerConfig

    def _get_topic_approximate_messages_count(self, ch: amqp.Channel, topic: str) -> int:
        res = ch.queue_declare(queue=topic, passive=True)
        self.echo.log_debug(f'topic "{topic}" has messages={res.message_count}, consumers={res.consumer_count}')
        return int(res.message_count)

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        with self.connected_connection.channel() as ch:
            return self._get_topic_approximate_messages_count(ch, topic)

    @classmethod
    def get_connection_uri(cls) -> str:
        raise NotImplementedError(f"cls method get_connection_uri must be implemented for class '{cls.__name__}'")

    @functools.cached_property
    def parsed_connection_uri(self) -> urllib.parse.ParseResult:
        return urlparse(url=self.get_connection_uri())

    def _connection_instance(self) -> amqp.Connection:
        return amqp.Connection(
            host=f'{self.parsed_connection_uri.hostname}:{self.parsed_connection_uri.port}',
            password=self.parsed_connection_uri.password,
            userid=self.parsed_connection_uri.username,
            heartbeat=self.config.heartbeat,
        )

    __slots__ = (
        '_consumers',
        '_connection',
        '_consuming_enabled',
        '_in_processing',
        '_interrupted',
        '_initialized_exchanges',
        '_initialized_topics',
        '_retry_consuming',
        '_heartbeat_enabled',
        '_heartbeat_delay',
        '_heartbeat_thread',
    )

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        super().__init__(mediator, definition)

        self._consumers: List[UniAmqpPyBrokerConsumer] = list()

        self._connection = self._connection_instance()

        self._consuming_enabled = False
        self._in_processing = False
        self._interrupted = False

        self._initialized_exchanges: Set[str] = set()
        self._initialized_topics: Set[str] = set()

        def raise_connection_err() -> None:
            raise ConnectionError()

        self._publish = retryable(  # type: ignore
            self._publish,
            self.echo.mk_child('publishing'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )
        self.get_topic_approximate_messages_count = retryable(  # type: ignore
            self.get_topic_approximate_messages_count,
            self.echo.mk_child('get_topic_approximate_messages_count'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )
        self.rpc_call = retryable(  # type: ignore
            self.rpc_call,
            self.echo.mk_child('rpc_call'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )
        self._retry_consuming = retryable(
            self._consuming,
            self.echo.mk_child('consuming'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )
        self._ack = retryable(  # type: ignore
            self._ack,
            self.echo.mk_child('ack'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )
        self._reject = retryable(  # type: ignore
            self._reject,
            self.echo.mk_child('reject'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )
        self.publish_answer = retryable(  # type: ignore
            self.publish_answer,
            self.echo.mk_child('publish_answer'),
            retry_max_count=self.config.retry_max_count,
            retry_delay_s=self.config.retry_delay_s,
            retryable_errors=RECOVERABLE_ERRORS,
            on_retries_ends=raise_connection_err,
        )

        self._heartbeat_enabled = False
        self._heartbeat_delay = self.config.heartbeat / 2
        self._heartbeat_thread: threading.Thread = threading.Thread(
            name=f'broker-{self.definition.name}-heartbeat',
            target=self._heartbeat_tick_loop,
            daemon=False,
            kwargs=dict(
                delay_s=self._heartbeat_delay,
            )
        )

    @property
    def connected_connection(self) -> amqp.Connection:
        self.connect()
        return self._connection

    def _init_topic(self, ch: amqp.Channel, exchange: str, topic: str) -> str:
        q_key = f'{exchange}->{topic}'
        if q_key in self._initialized_topics:
            return topic
        self._initialized_topics.add(q_key)

        if exchange not in self._initialized_topics:
            self._initialized_exchanges.add(exchange)
            ch.exchange_declare(
                exchange=exchange,
                type="direct",
                passive=False,
                durable=True,
                auto_delete=False,
            )
            self.echo.log_info(f'exchange "{exchange}" initialized')

        if exchange == self.config.exchange_name:
            ch.queue_declare(queue=topic, durable=True, auto_delete=False, passive=False)
        elif exchange == self.config.answer_exchange_name:
            ch.queue_declare(queue=topic, durable=False, auto_delete=True, exclusive=True, passive=False)
        else:
            raise TypeError(f'invalid exchange name "{exchange}"')

        ch.queue_bind(queue=topic, exchange=exchange, routing_key=topic)
        self.echo.log_info(f'queue "{q_key}" initialized')
        return topic

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        return

    def stop_consuming(self) -> None:
        if not self._consuming_enabled:
            return
        self._interrupted = True
        if not self._in_processing:
            # self.close()
            self._consuming_enabled = False
            self.echo.log_info('consumption stopped')

    def connect(self) -> None:
        try:
            if not self._connection.connected:
                self._connection = self._connection_instance()
                self._connection.connect()
                self._connection.send_heartbeat()
                self.echo.log_info('connected')
        except RECOVERABLE_ERRORS as e:
            self.echo.log_error(f'connection problem :: {e}')
            raise ConnectionError(str(e))

    def close(self) -> None:
        self.echo.log_info('closing')
        self._stop_heartbeat_tick()

        if not self._connection.connected or self._connection.is_closing:
            self.echo.log_info('already closed')
            return

        try:
            self._connection.close()
        except AMQPError:
            pass
        self.echo.log_info('closed')

    def _ack(self, ch: amqp.Channel, delivery_tag: str) -> None:
        if not ch.is_open or ch.is_closing:
            ch = self.connected_connection.channel()
        ch.basic_ack(delivery_tag=delivery_tag)
        self.echo.log_info(f'message "{delivery_tag}" ACK')

    def _reject(self, ch: amqp.Channel, delivery_tag: str) -> None:
        if not ch.is_open or ch.is_closing:
            ch = self.connected_connection.channel()
        ch.basic_reject(delivery_tag=delivery_tag, requeue=True)
        self.echo.log_warning(f'message "{delivery_tag}" REJECT')

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        echo = self.echo.mk_child(f'topic[{consumer.topic}]')
        if self._consuming_enabled:
            raise OverflowError(f'you cannot add consumer dynamically :: tag="{consumer.id}" group_id={consumer.group_id}')

        def consumer_wrapper(ch: amqp.Channel, message: amqp.Message) -> None:
            self.echo.log_info(f'message "{message.delivery_tag}" received from topic "{consumer.topic}" (in consumer {consumer.id})')
            self._in_processing = True

            meta = self.parse_message_body(
                message.body,
                compression=message.headers.get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                content_type=message.content_type,
                unwrapped=consumer.unwrapped,
            )

            manager = UniPyPikaBrokerMessageManager(ch, message.delivery_tag, reject=self._reject, ack=self._ack)
            consumer.message_handler(meta, manager)

            self._in_processing = False
            if self._interrupted:
                self.stop_consuming()

        self._consumers.append(UniAmqpPyBrokerConsumer(
            queue=consumer.topic,
            on_message_callback=consumer_wrapper,
            consumer_tag=consumer.id,
            prefetch_count=consumer.prefetch_count,
        ))

        echo.log_info(f'added consumer :: tag="{consumer.id}" group_id={consumer.group_id}')

    def _stop_heartbeat_tick(self) -> None:
        if not self._heartbeat_enabled:
            return
        self._heartbeat_enabled = False
        if self._heartbeat_thread.is_alive():
            self._heartbeat_thread.join()

    def _start_heartbeat_tick(self) -> None:
        if self._heartbeat_enabled:
            return
        self._heartbeat_enabled = self._heartbeat_delay > 0

        if not self._heartbeat_enabled:
            self.echo.log_info('heartbeat disabled')
            return

        self.echo.log_info(f'heartbeat enabled :: delay {self._heartbeat_delay}s')
        self._heartbeat_thread.start()

    def _heartbeat_tick_loop(self, delay_s: float, ping_rate: int = 10) -> None:
        delay_part_s = delay_s / ping_rate
        counter = 0
        while self._heartbeat_enabled:
            counter += 1
            sleep(delay_part_s)
            if counter != ping_rate:  # each ping_rate-1
                continue
            counter = 0
            try:
                self._connection.heartbeat_tick()
                self.echo.log_debug('heartbeat tick')
            except ConnectionForced as e:
                self.echo.log_error(str(e))
                self._heartbeat_enabled = False
                return
            except RECOVERABLE_ERRORS as e:
                self.echo.log_warning(f'ignored heartbeat error :: {e}')
                continue
            except AMQPError as e:
                self.echo.log_error(str(e))
                self._heartbeat_enabled = False
                return

    def _consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if len(self._consumers) == 0:
            echo.log_warning('has no consumers to start consuming')
            return

        self.connect()

        for c in self._consumers:
            ch = self._connection.channel()
            topic = self._init_topic(ch, self.config.exchange_name, c.queue)
            ch.basic_consume(queue=topic, callback=functools.partial(c.on_message_callback, ch), consumer_tag=c.consumer_tag)
            ch.basic_qos(prefetch_count=self.config.prefetch, a_global=False, prefetch_size=0)
            echo.log_debug(f'added consumer {c.consumer_tag} on {self.config.exchange_name}->{topic}')

        echo.log_info(f'starting consuming :: consumers_count={len(self._consumers)}')

        self._start_heartbeat_tick()
        while self._consuming_enabled:
            echo.log_debug('wait for next message ...')
            self._connection.drain_events()
        self._stop_heartbeat_tick()

    def start_consuming(self) -> None:
        if self._consuming_enabled:
            return
        self._consuming_enabled = True
        self._interrupted = False
        self._in_processing = False

        self._retry_consuming()

    def _alone(self, ch: amqp.Channel, topic: str, alone: bool) -> bool:
        if alone:
            size = self._get_topic_approximate_messages_count(ch, topic)
            if size > 0:
                self.echo.log_info(f'sending was skipped, because topic {topic} has messages: {size}>0')
                return True
        return False

    def _publish_ch(self, ch: amqp.Channel, exchange: str, topic: str, meta: UniMessageMeta, props: UniAmqpPyBrokerMsgProps, alone: bool = False) -> None:
        if self._alone(ch, topic, alone):
            return

        self.echo.log_debug(f'message start publishing to {exchange}->{topic}')
        topic = self._init_topic(ch, exchange, topic)
        ch.basic_publish(
            amqp.Message(body=self.serialize_message_body(meta), **props._asdict()),
            exchange=exchange,
            routing_key=topic,
            mandatory=self.config.mandatory_publishing,
            # immediate=self.config.immediate_publishing,
        )
        self.echo.log_debug(f'message published to {exchange}->{topic}')

    def _publish(self, exchange: str, topic: str, meta: UniMessageMeta, props: UniAmqpPyBrokerMsgProps, alone: bool = False) -> None:
        with self.connected_connection.channel() as ch:
            return self._publish_ch(ch, exchange, topic, meta, props, alone)

    def _mk_mesg_props_by_meta(self, meta: UniMessageMeta) -> UniAmqpPyBrokerMsgProps:
        headers = dict()
        if self.definition.compression is not None:
            headers[BASIC_PROPERTIES__HEADER__COMPRESSION_KEY] = self.definition.compression
        if meta.ttl_s:
            headers['x-message-ttl'] = str(meta.ttl_s * 1000)

        if meta.need_answer:
            assert meta.answer_params is not None
            return UniAmqpPyBrokerMsgProps(
                content_type=self.definition.content_type,
                content_encoding='utf-8',
                reply_to=f'{meta.answer_params.topic}.{meta.answer_params.id}',
                correlation_id=str(meta.id),
                delivery_mode=2 if self.config.persistent_message else 0,
                application_headers=headers,
            )
        return UniAmqpPyBrokerMsgProps(
            content_type=self.definition.content_type,
            content_encoding='utf-8',
            delivery_mode=2 if self.config.persistent_message else 0,
            application_headers=headers,
        )

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        for meta in meta_list:  # TODO: package sending
            props = self._mk_mesg_props_by_meta(meta)
            self._publish(self.config.exchange_name, topic, meta, props, alone)
        self.echo.log_info(f'{list(meta_list)} messages published to {self.config.exchange_name}->{topic}')

    def rpc_call(self, topic: str, meta: UniMessageMeta, *, alone: bool = False, max_delay_s: int = 1, unwrapped: bool = False) -> Optional[UniMessageMeta]:
        assert meta.answer_params is not None
        with self.connected_connection.channel() as ch:
            if alone:
                size = self._get_topic_approximate_messages_count(ch, topic)
                if size > 0:
                    self.echo.log_info(f'sending was skipped, because topic {topic} has messages: {size}>0')
                    return None

            answ_topic = self._init_topic(ch, self.config.answer_exchange_name, f'{meta.answer_params.topic}.{meta.answer_params.id}')
            self._publish_ch(ch, self.config.exchange_name, topic, meta, self._mk_mesg_props_by_meta(meta), alone)

            started = time.time()
            self.echo.log_info(f'waiting for message from {self.config.answer_exchange_name}->{answ_topic}')
            while True:
                msg: Optional[amqp.Message] = ch.basic_get(queue=answ_topic)

                if msg is None:
                    if (time.time() - started) > max_delay_s:
                        raise UniAnswerDelayError(f'answer for {self.config.answer_exchange_name}->{answ_topic} reached delay limit {max_delay_s} seconds')
                    self.echo.log_debug(f'no answer {int(time.time() - started + 1)}s in {self.config.answer_exchange_name}->{answ_topic}')
                    sleep(0.1)
                    continue

                self.echo.log_debug(f'got answer from {self.config.answer_exchange_name}->{answ_topic}')

                return self.parse_message_body(
                    msg.body,
                    compression=(msg.headers or dict()).get(BASIC_PROPERTIES__HEADER__COMPRESSION_KEY, None),
                    content_type=msg.content_type,
                    unwrapped=unwrapped,
                )

    def publish_answer(self, answer_params: UniAnswerParams, meta: UniMessageMeta) -> None:
        props = UniAmqpPyBrokerMsgProps(
            content_type=self.definition.content_type,
            content_encoding='utf-8',
            delivery_mode=1,
        )
        with self.connected_connection.channel() as ch:
            self._publish_ch(ch, self.config.answer_exchange_name, f'{answer_params.topic}.{answer_params.id}', meta, props, alone=False)



from typing import Tuple

from unipipeline.definitions.uni_dynamic_definition import UniDynamicDefinition


class UniKafkaBrokerConfig(UniDynamicDefinition):
    api_version: Tuple[int, ...]
    retry_max_count: int = 100
    retry_delay_s: int = 3



from datetime import datetime
from typing import NamedTuple, List, Tuple, Optional, Iterable, TYPE_CHECKING

from crontab import CronTab  # type: ignore

from unipipeline.message.uni_cron_message import UniCronMessage
from unipipeline.definitions.uni_cron_task_definition import UniCronTaskDefinition

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


class UniCronJob(NamedTuple):
    id: int
    task: UniCronTaskDefinition
    crontab: Optional[CronTab]
    every_sec: Optional[int]
    mediator: 'UniMediator'
    message: UniCronMessage

    def next_sec(self) -> int:
        if self.crontab is not None:
            return int(self.crontab.next(default_utc=False))
        sec = datetime.now().second
        mod = sec % self.every_sec  # type: ignore
        return 0 if mod == 0 else (self.every_sec - mod)  # type: ignore

    @staticmethod
    def mk_jobs_list(tasks: Iterable[UniCronTaskDefinition], mediator: 'UniMediator') -> List['UniCronJob']:
        res = list()
        for i, task in enumerate(tasks):
            res.append(UniCronJob(
                id=i,
                task=task,
                crontab=CronTab(task.when) if task.when is not None else None,
                every_sec=task.every_sec,
                mediator=mediator,
                message=UniCronMessage(task_name=task.name)
            ))
        return res

    @staticmethod
    def search_next_tasks(all_tasks: List['UniCronJob']) -> Tuple[Optional[int], List['UniCronJob']]:
        min_delay: Optional[int] = None
        notification_list: List[UniCronJob] = []
        for cj in all_tasks:
            sec = cj.next_sec()
            if min_delay is None:
                min_delay = sec
            if sec < min_delay:
                notification_list.clear()
                min_delay = sec
            if sec <= min_delay:
                notification_list.append(cj)

        return min_delay, notification_list

    def send(self) -> None:
        self.mediator.send_to(self.task.worker.name, self.message, alone=self.task.alone)



from typing import Optional, Any, Dict, List, Set, TYPE_CHECKING

from kafka import KafkaProducer, KafkaConsumer  # type: ignore

from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.brokers.uni_kafka_broker_config import UniKafkaBrokerConfig
from unipipeline.brokers.uni_kafka_broker_message_manager import UniKafkaBrokerMessageManager
from unipipeline.message_meta.uni_message_meta import UniMessageMeta

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


class UniKafkaBroker(UniBroker[UniKafkaBrokerConfig]):
    config_type = UniKafkaBrokerConfig

    def get_boostrap_servers(self) -> List[str]:
        raise NotImplementedError(f'method get_boostrap_server must be implemented for {type(self).__name__}')

    def get_security_conf(self) -> Dict[str, Any]:
        raise NotImplementedError(f'method get_security_conf must be implemented for {type(self).__name__}')

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        super().__init__(mediator, definition)

        self._bootstrap_servers = self.get_boostrap_servers()

        self._producer: Optional[KafkaProducer] = None

        self._security_conf: Dict[str, Any] = self.get_security_conf()

        self._consumers: List[UniBrokerConsumer] = list()
        self._kfk_active_consumers: List[KafkaConsumer] = list()

        self._consuming_started = False
        self._interrupted = False
        self._in_processing = False

    def stop_consuming(self) -> None:    # TODO
        self._end_consuming()

    def _end_consuming(self) -> None:
        if not self._consuming_started:
            return
        self._interrupted = True
        if not self._in_processing:
            for kfk_consumer in self._kfk_active_consumers:
                kfk_consumer.close()
            self._consuming_started = False
            self.echo.log_info('consumption stopped')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        return 0  # TODO

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        pass  # TODO

    def connect(self) -> None:
        if self._producer is not None:
            if self._producer._closed:
                self._producer.close()
                self._producer = None
            else:
                return

        # TODO: change default connection as producer yo abstract connection to kafka server
        self._producer = KafkaProducer(
            bootstrap_servers=self._bootstrap_servers,
            api_version=self.config.api_version,
            retries=self.config.retry_max_count,
            acks=1,
            **self._security_conf,
        )

        if not self._producer.bootstrap_connected():
            raise ConnectionError()

        self.echo.log_info('connected')

    def close(self) -> None:
        if self._producer is not None:
            self._producer.close()
            self._producer = None
        for kfk_consumer in self._kfk_active_consumers:
            kfk_consumer.close()

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        self._consumers.append(consumer)

    def start_consuming(self) -> None:
        echo = self.echo.mk_child('consuming')
        if len(self._consumers) == 0:
            echo.log_warning('has no consumers to start consuming')
            return
        if self._consuming_started:
            echo.log_warning('consuming has already started. ignored')
            return
        self._consuming_started = True
        self._interrupted = False
        self._in_processing = False

        if len(self._consumers) != 1:
            raise OverflowError('invalid consumers number. this type of brokers not supports multiple consumers')

        consumer = self._consumers[0]
        kfk_consumer = KafkaConsumer(
            consumer.topic,
            api_version=self.config.api_version,
            bootstrap_servers=self._bootstrap_servers,
            enable_auto_commit=False,
            group_id=consumer.group_id,
        )

        self._kfk_active_consumers.append(kfk_consumer)

        def commit() -> None:
            kfk_consumer.commit()

        # TODO: retry
        for consumer_record in kfk_consumer:
            self._in_processing = True

            meta = self.parse_message_body(consumer_record.value, self.definition.compression, self.definition.content_type, consumer.unwrapped)

            manager = UniKafkaBrokerMessageManager(commit)
            consumer.message_handler(meta, manager)

            self._in_processing = False
            if self._interrupted:
                self._end_consuming()
                break

        for kfk_consumer in self._kfk_active_consumers:
            kfk_consumer.close()

    def _get_producer(self) -> KafkaProducer:
        self.connect()
        assert self._producer is not None
        return self._producer

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        # TODO: alone
        # TODO: ttl
        # TODO: retry
        self.echo.log_debug(f'publishing the messages: {meta_list}')

        p = self._get_producer()

        for meta in meta_list:
            p.send(
                topic=topic,
                value=self.serialize_message_body(meta),
                key=str(meta.id).encode('utf8')
            )
        p.flush()



from typing import Dict, TypeVar, Set, List, TYPE_CHECKING
from uuid import UUID

from unipipeline.brokers.uni_broker import UniBroker
from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.brokers.uni_memory_broker_queue import UniMemoryBrokerQueue
from unipipeline.definitions.uni_definition import UniDynamicDefinition
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniAnswerParams

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


TItem = TypeVar('TItem')


class UniMemoryBroker(UniBroker[UniDynamicDefinition]):

    def stop_consuming(self) -> None:
        pass

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        super(UniMemoryBroker, self).__init__(mediator, definition)
        self._consuming_started = False
        self._queues_by_topic: Dict[str, UniMemoryBrokerQueue] = dict()
        self._consumers_count = 0

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        return self._queues_by_topic[topic].messages_to_process_count()

    def initialize(self, topics: Set[str], answer_topic: Set[str]) -> None:
        for topic in topics:
            self._init_queue(topic)

    def _init_queue(self, topic: str) -> None:
        if topic in self._queues_by_topic:
            return
        self._queues_by_topic[topic] = UniMemoryBrokerQueue(self.echo.mk_child(f'topic[{topic}]'))

    def _get_answer_topic_name(self, topic: str, answ_id: UUID) -> str:
        return f'answer@{topic}@{answ_id}'

    def connect(self) -> None:
        pass

    def close(self) -> None:
        pass

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        self._consumers_count += 1
        self._queues_by_topic[consumer.topic].add_listener(consumer.message_handler, 1)
        self.echo.log_info(f'consumer for topic "{consumer.topic}" added with consumer_tag "{consumer.id}"')

    def start_consuming(self) -> None:
        if self._consuming_started:
            raise OverflowError('consuming has already started')

        if self._consumers_count == 0:
            self.echo.log_warning('has no consumers')
            return

        self._consuming_started = True

        self.echo.log_info('start consuming')
        for ql in self._queues_by_topic.values():
            ql.process_all()

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        ql = self._queues_by_topic[topic]
        for meta in meta_list:
            ql.add(meta)
        if self._consuming_started:
            ql.process_all()

    def get_answer(self, answer_params: UniAnswerParams, max_delay_s: int, unwrapped: bool) -> UniMessageMeta:
        topic = self._get_answer_topic_name(answer_params.topic, answer_params.id)
        self._init_queue(topic)

        answ_meta = self._queues_by_topic[topic].get_next()

        if unwrapped:
            return UniMessageMeta.create_new(answ_meta.payload, unwrapped=True)

        return answ_meta

    def publish_answer(self, answer_params: UniAnswerParams, meta: UniMessageMeta) -> None:
        topic = self._get_answer_topic_name(answer_params.topic, answer_params.id)
        self._init_queue(topic)
        self._queues_by_topic[topic].add(meta)



import os.path
import sys
from time import sleep

CWD = os.path.dirname(os.path.abspath(__file__))

sys.path.append(os.path.dirname(CWD))

from unipipeline.modules.uni import Uni
from example.args import args

u = Uni(f"{CWD}/dag-{args.type}.yml")

u.init_producer_worker('input_worker')

u.initialize()

for i in range(args.produce_count):
    u.send_to("input_worker", dict(value=i, some="Привет World"))
    sleep(1)

# u.exit()



from example.messages.ender_after_cron_answer_message import EnderAfterCronAnswerMessage
from example.messages.ender_after_cron_input_message import EnderAfterCronInputMessage
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage


class EnderAfterCronWorker(UniWorker[EnderAfterCronInputMessage, EnderAfterCronAnswerMessage]):
    def handle_message(self, msg: UniWorkerConsumerMessage[EnderAfterCronInputMessage]) -> EnderAfterCronAnswerMessage:
        return EnderAfterCronAnswerMessage(
            value=f'after_cron => {msg.payload.value}'
        )



# from example.messages.ender_after_cron_answer_message import EnderAfterCronAnswerMessage
# from example.messages.ender_after_cron_input_message import EnderAfterCronInputMessage
# from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.message.uni_cron_message import UniCronMessage
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage


class MySuperCronWorker(UniWorker[UniCronMessage, None]):
    def handle_message(self, msg: UniWorkerConsumerMessage[UniCronMessage]) -> None:
        print(f'!!!!! MySuperCronWorker MESSAGE {msg.payload}')  # noqa
        # answ: UniAnswerMessage[EnderAfterCronAnswerMessage] = self.manager.get_answer_from('ender_after_cron_worker', EnderAfterCronInputMessage(value=f'cron>>>{msg.payload.task_name}'))
        #
        # print(answ.payload.value)  # noqa



from typing import Generic, TypeVar, Dict, Any, Set

TConnectionObj = TypeVar('TConnectionObj')


class ConnectionObj(Generic[TConnectionObj]):
    def __hash__(self) -> int:
        raise NotImplementedError(f'method "__hash__" is not implemented for {type(self).__name__}')

    def get(self) -> TConnectionObj:
        raise NotImplementedError(f'method "get" is not implemented for {type(self).__name__}')

    def is_closed(self) -> bool:
        raise NotImplementedError(f'method "is_closed" is not implemented for {type(self).__name__}')

    def connect(self) -> None:
        raise NotImplementedError(f'method "connect" is not implemented for {type(self).__name__}')

    def close(self) -> None:
        raise NotImplementedError(f'method "close" is not implemented for {type(self).__name__}')


class ConnectionRC(Generic[TConnectionObj]):
    def __init__(self, connection_obj: ConnectionObj[TConnectionObj]) -> None:
        self._connection_obj = connection_obj
        self._identifiers: Set[int] = set()

    def count(self) -> int:
        return len(self._identifiers)

    def connect(self, identifier: int) -> TConnectionObj:
        if identifier not in self._identifiers:
            if self.count() == 0:
                self._connection_obj.connect()
            self._identifiers.add(identifier)
        return self._connection_obj.get()

    def disconnect(self, identifier: int) -> None:
        if identifier not in self._identifiers:
            return
        self._identifiers.remove(identifier)
        if self.count() == 0:
            self._connection_obj.close()


class ConnectionManager(Generic[TConnectionObj]):
    def __init__(self, rc: ConnectionRC[TConnectionObj], identifier: int, connection_obj: ConnectionObj[Any]) -> None:
        self._rc = rc
        self._id = identifier
        self._key = hash(connection_obj)
        self._connection_obj = connection_obj

    def connect(self) -> TConnectionObj:
        return self._rc.connect(self._id)

    def disconnect(self) -> None:
        return self._rc.disconnect(self._id)

    def close(self) -> None:
        self.disconnect()

    def __del__(self) -> None:
        self.disconnect()

    def __enter__(self) -> TConnectionObj:
        return self.connect()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return self.disconnect()


class ConnectionPool:

    def __init__(self) -> None:
        self._registry: Dict[int, ConnectionRC[Any]] = dict()
        self._id_counter: int = 0

    def new_manager(self, connection_obj: ConnectionObj[TConnectionObj]) -> ConnectionManager[TConnectionObj]:
        identifier = self._id_counter
        self._id_counter += 1

        key = hash(connection_obj)
        if key not in self._registry:
            self._registry[key] = ConnectionRC(connection_obj)
        rc = self._registry[key]

        return ConnectionManager(rc, identifier, connection_obj)


connection_pool = ConnectionPool()



from setuptools import setup, find_packages

from codecs import open
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="unipipeline",
    version="1.6.0",
    description="simple way to build the declarative and distributed data pipelines with python. it supports rabbitmq or kafka as a broker",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aliaksandr-master/unipipeline",
    author="Aliaksandr Master",
    author_email="alxe.master@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            'unipipeline=unipipeline.main:main',
        ],
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "jinja2>=3.0.3",
        "amqp>=5.0.6",
        "pyyaml>=5.4.12",
        "types-pyyaml>=5.4.12",
        "pydantic>=1.8.2",
        "psycopg2-binary>=2.9.2",
        "pika>=1.2.0",
        "crontab>=0.23.0",
        "kafka-python>=2.0.2",
    ],
)
